import pulumi
from pulumi_aws import rds
from pulumi.output import Output

from pydantic.dataclasses import dataclass, ConfigDict
from typing import Optional, List, Mapping, Literal

from theia import ResourceOptions
from theia.utils.logger import Logger
from theia.registry import StackRegistry
from theia.definitions import THEIA_VERSION
from theia.utils.errors import TheiaConfigurationError

logger = Logger("rds")


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class ClusterSettings:
    availability_zones: Optional[List[str]] = None
    backup_retention_period: Optional[int] = 30
    copy_tags_to_snapshot: Optional[bool] = True
    deletion_protection: Optional[bool] = False
    enable_global_write_forwarding: Optional[bool] = False
    enabled_cloudwatch_logs_exports: Optional[
        List[Literal["audit", "error", "general", "slowquery", "postgresql"]]
    ] = None

    # TODO Multi az check for mysql/postgres
    engine: Optional[
        Literal["aurora", "aurora-mysql", "aurora-postgresql", "mysql", "postgres"]
    ] = "aurora"  # Multi az
    engine_version: Optional[str] = None
    final_snapshot_identifier: Optional[str] = None
    global_cluster_identifier: Optional[str] = None
    preferred_backup_window: Optional[str] = None
    preferred_maintenance_window: Optional[str] = None
    vpc_security_group_ids: Optional[List[str | Output[str]]] = None

    # TODO Better multi az implementation
    # Multi AZ
    allocated_storage: Optional[int] = None
    db_cluster_instance_class: Optional[str] = None
    iops: Optional[int] = None
    storage_type: Optional[str] = None

    def __post_init__(self):
        if (
            (self.availability_zones and len(self.availability_zones) > 1)
            and (not self.iops or not self.db_cluster_instance_class)
            and (not self.storage_type or not self.allocated_storage)
        ):
            raise TheiaConfigurationError(
                "iops, db_cluster_instance_class, allocated_storage and storage type are required for multi AZ cluster"
            )

        if self.enable_global_write_forwarding and not self.global_cluster_identifier:
            raise TheiaConfigurationError(
                "global write forwarding is enabled but global cluster identifier was not provided"
            )


@dataclass
class ClusterInstanceArgs:
    instance: str
    auto_minor_version_upgrade: Optional[bool] = False
    publicly_accessible: Optional[bool] = False
    availability_zone: Optional[str] = None
    copy_tags_to_snapshot: Optional[bool] = True
    apply_immediately: Optional[bool] = False


@dataclass
class ParameterArg:
    name: str
    value: str


@dataclass
class ParameterGroup:
    description: str
    params: List[ParameterArg]


class Cluster(pulumi.ComponentResource):
    def __init__(
        self,
        resource_name: str,
        database_name: str,
        master_username: str,
        master_password: str,
        family: str,
        cluster_parameter_groups: Optional[ParameterGroup] = None,
        db_parameter_groups: Optional[ParameterGroup] = None,
        subnet_ids: Optional[List[str]] = None,
        instances: Optional[List[ClusterInstanceArgs]] = None,
        settings: Optional[ClusterSettings] = ClusterSettings(),
        tags: Optional[Mapping[str, str]] = None,
        opts: Optional[ResourceOptions] = None,
    ) -> None:
        super().__init__("theia:aws:database", resource_name, None, opts)

        merged_tags = THEIA_VERSION
        name = {"Name": resource_name}
        if tags:
            merged_tags = {**tags, **name, **merged_tags}

        if cluster_parameter_groups:
            logger.info(
                f"creating resource cluster parameter group for {resource_name}"
            )
            parameters = []
            for param in cluster_parameter_groups.params:
                parameters.append(
                    rds.ClusterParameterGroupParameterArgs(
                        name=param.name,
                        value=param.value,
                    )
                )

            self.cluster_pg = rds.ClusterParameterGroup(
                f"{resource_name}-cluster-pg",
                description=cluster_parameter_groups.description,
                family=family,
                name=f"{resource_name}-cluster-pg",
                parameters=parameters,
                opts=ResourceOptions(parent=self),
                tags=merged_tags,
            )
        else:
            logger.info(
                f"creating resource default cluster parameter group for {resource_name}"
            )
            parameters = [
                rds.ClusterParameterGroupParameterArgs(
                    name="max_allowed_packet", value="268435456"
                ),
                rds.ClusterParameterGroupParameterArgs(
                    name="slow_query_log", value="1"
                ),
                rds.ClusterParameterGroupParameterArgs(
                    name="log_queries_not_using_indexes", value="1"
                ),
                rds.ClusterParameterGroupParameterArgs(
                    name="long_query_time", value="1"
                ),
            ]
            self.cluster_pg = rds.ClusterParameterGroup(
                f"{resource_name}-default-cluster-pg",
                description=f"default parameter group for {resource_name}",
                family=family,
                name=f"{resource_name}-default-cluster-pg",
                parameters=parameters,
                opts=ResourceOptions(parent=self),
                tags=merged_tags,
            )

        if subnet_ids:
            logger.info(f"creating resource subnet group for {resource_name}")
            self.subnet_group = rds.SubnetGroup(
                f"{resource_name}-subnet-group",
                name=f"{resource_name}-subnet-group",
                subnet_ids=subnet_ids,
                description=f"Subnet Group for {resource_name}",
                opts=ResourceOptions(parent=self),
                tags=merged_tags,
            )

        if db_parameter_groups:
            logger.info(f"creating resource db parameter group for {resource_name}")
            parameters = []
            for param in db_parameter_groups.params:
                parameters.append(
                    rds.ParameterGroupParameterArgs(name=param.name, value=param.value)
                )

            self.db_pg = rds.ParameterGroup(
                f"{resource_name}-db-pg",
                description=db_parameter_groups.description,
                family=family,
                name=f"{resource_name}-db-pg",
                parameters=parameters,
                opts=ResourceOptions(parent=self),
                tags=merged_tags,
            )

        cluster_args = rds.ClusterArgs(
            allocated_storage=settings.allocated_storage,
            availability_zones=settings.availability_zones,
            backup_retention_period=settings.backup_retention_period,
            cluster_identifier=f"{resource_name}-cluster",
            copy_tags_to_snapshot=settings.copy_tags_to_snapshot,
            database_name=database_name,
            db_cluster_instance_class=settings.db_cluster_instance_class,
            db_cluster_parameter_group_name=self.cluster_pg.name,
            db_subnet_group_name=None if not subnet_ids else self.subnet_group.name,
            deletion_protection=settings.deletion_protection,
            enable_global_write_forwarding=settings.enable_global_write_forwarding,
            enabled_cloudwatch_logs_exports=settings.enabled_cloudwatch_logs_exports,
            engine=settings.engine,
            engine_version=settings.engine_version,
            final_snapshot_identifier=settings.final_snapshot_identifier,
            global_cluster_identifier=settings.global_cluster_identifier,
            master_password=master_password,
            master_username=master_username,
            preferred_backup_window=settings.preferred_backup_window,
            preferred_maintenance_window=settings.preferred_maintenance_window,
            skip_final_snapshot=True
            if not settings.final_snapshot_identifier
            else False,
            tags=merged_tags,
            vpc_security_group_ids=settings.vpc_security_group_ids,
        )
        depends = []
        depends.append(self.cluster_pg)
        if subnet_ids:
            depends.append(self.subnet_group)

        logger.info(f"creating rds cluster resource {resource_name}")
        self.rds = rds.Cluster(
            resource_name=f"{resource_name}-cluster",
            args=cluster_args,
            opts=ResourceOptions(parent=self, depends_on=depends),
        )

        if instances:
            self.instances = []
            logger.info(f"creating rds instance resource for {resource_name}")
            for i, instance in enumerate(instances):
                self.instances.append(
                    rds.ClusterInstance(
                        f"{resource_name}-instance-{i}",
                        identifier=f"{resource_name}-instance-{i}",
                        instance_class=instance.instance,
                        cluster_identifier=self.rds.cluster_identifier,
                        auto_minor_version_upgrade=instance.auto_minor_version_upgrade,
                        db_subnet_group_name=None
                        if not subnet_ids
                        else self.subnet_group.name,
                        db_parameter_group_name=None
                        if not db_parameter_groups
                        else self.db_pg.name,
                        engine=self.rds.engine,
                        engine_version=self.rds.engine_version,
                        apply_immediately=instance.apply_immediately,
                        publicly_accessible=instance.publicly_accessible,
                        copy_tags_to_snapshot=instance.copy_tags_to_snapshot,
                        availability_zone=instance.availability_zone,
                        tags=merged_tags,
                        opts=ResourceOptions(parent=self.rds, depends_on=[self.rds]),
                    )
                )

        pulumi.export(
            f"rds---{resource_name}",
            {
                "id": self.rds.id,
                "arn": self.rds.arn,
                "cluster_resource_id": self.rds.cluster_resource_id,
                "endpoint": self.rds.endpoint,
                "hosted_zone_id": self.rds.hosted_zone_id,
            },
        )
        self.register_outputs({})

    @staticmethod
    def import_(stack: str, resource_name: str) -> rds.Cluster:
        stack_ref = StackRegistry.get(stack_name=stack)
        rds = stack_ref.get_output(name=f"rds---{resource_name}")
        return rds.Cluster.get(resource_name=resource_name, id=rds["id"])
