import pulumi

from pulumi_aws import ec2
from pulumi.output import Output

from typing import Optional, Mapping, Literal
from pydantic.dataclasses import dataclass, ConfigDict

from theia import ResourceOptions
from theia.utils.logger import Logger
from theia.registry import StackRegistry
from theia.definitions import THEIA_VERSION
from theia.aws.network.eip import Eip

logger = Logger("ngw")


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class NatGatewaySettings:
    connectivity_type: Optional[Literal["public", "private"]] = "public"
    private_ip: Optional[str | Output[str]] = None


class NatGateway:
    def __new__(
        cls,
        resource_name: str,
        subnet_id: str | Output[str],
        tags: Optional[Mapping[str, str]] = None,
        settings: Optional[NatGatewaySettings] = NatGatewaySettings(),
        opts: Optional[ResourceOptions] = None,
    ) -> ec2.NatGateway:
        merged_tags = THEIA_VERSION
        name = {"Name": resource_name}
        if tags:
            merged_tags = {**tags, **name, **merged_tags}

        if settings.connectivity_type == "public":
            eip = Eip(f"{resource_name}-eip", opts=opts)
            allocation_id = eip.allocation_id

        ngw_args = ec2.NatGatewayArgs(
            subnet_id=subnet_id,
            allocation_id=allocation_id,
            connectivity_type=settings.connectivity_type,
            private_ip=settings.private_ip,
            tags=merged_tags,
        )

        logger.info(f"creating nat gateway resource {resource_name}")
        ngw = ec2.NatGateway(resource_name, args=ngw_args, opts=opts)

        export_values = {
            "id": ngw.id,
            "public_ip": ngw.public_ip,
            "private_ip": ngw.private_ip,
            "network_interface_id": ngw.network_interface_id,
        }

        pulumi.export(f"nat_gateway---{resource_name}", export_values)

        return ngw

    @staticmethod
    def import_(stack: str, resource_name: str) -> ec2.NatGateway:
        stack_ref = StackRegistry.get(stack_name=stack)
        ngw = stack_ref.get_output(name=f"nat_gateway---{resource_name}")
        return ec2.NatGateway.get(resource_name=resource_name, id=ngw["id"])
