import os
import json

from typing import Callable
from pulumi import automation as auto, Output, Input, ResourceOptions, StackReference

from typing import TypeVar, Optional, List

from theia.registry import StackRegistry
from theia.utils.logger import Logger
from theia.definitions import AWS_REGION

U = TypeVar("U")
T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

logger = Logger("theia")


class Theia:
    def __init__(
        self,
        stack: str,
        project: str,
        region: AWS_REGION,
        program: Callable[[], None],
        backend_url: str = None,
        config_phrase: str = None,
        stack_imports: List[str] = None,
    ) -> None:
        logger.info("setting pulumi config phrase: %s", config_phrase)
        os.environ["PULUMI_CONFIG_PASSPHRASE"] = (
            config_phrase if config_phrase is not None else ""
        )
        if backend_url:
            logger.info("setting custom backend url")
            os.environ["PULUMI_BACKEND_URL"] = backend_url
        logger.info("initializing stack: %s", stack)

        def infra():
            if stack_imports:
                for stack_import in stack_imports:
                    temp_stack_object = StackReference(name=stack_import)
                    StackRegistry.register(
                        stack_name=stack_import, stack_object=temp_stack_object
                    )
            program()

        self.stack = auto.create_or_select_stack(
            stack_name=stack,
            project_name=project,
            program=infra,
            opts=None
            if not backend_url
            else auto.LocalWorkspaceOptions(
                project_settings=auto.ProjectSettings(
                    name=project,
                    runtime="python",
                    backend=auto.ProjectBackend(backend_url),
                )
            ),
        )
        logger.info(f"settings stack config region: {region}")
        self.stack.set_config("aws:region", auto.ConfigValue(value=region))

    def preview(self, detailed_diff: bool = False, debug: bool = False) -> None:
        logger.info(f"previewing stack {self.stack.name}...")
        res = self.stack.preview(
            on_output=print, color="always", diff=detailed_diff, debug=debug
        )
        logger.info(f"preview summary: \n{json.dumps(res.change_summary, indent=2)}")

    def up(
        self,
        detailed_diff: bool = False,
        debug: bool = False,
        show_secrets: bool = False,
    ) -> None:
        logger.info(f"updating stack {self.stack.name}...")
        res = self.stack.up(
            on_output=print,
            color="always",
            diff=detailed_diff,
            debug=debug,
            show_secrets=show_secrets,
        )
        logger.info(
            f"update summary: \n{json.dumps(res.summary.resource_changes, indent=2)}"
        )

    def destroy(self, debug: bool = False) -> None:
        logger.info(f"destroying stack {self.stack.name}...")
        res = self.stack.destroy(on_output=print, color="always", debug=debug)
        logger.info(f"destroy summary: \n{json.dumps(res.summary.result, indent=2)}")

    def refresh(self, debug: bool = False) -> None:
        logger.info(f"refreshing stack {self.stack.name}...")
        res = self.stack.refresh(on_output=print, color="always", debug=debug)
        logger.info(
            f"refresh summary: \n{json.dumps(res.summary.resource_changes, indent=2)}"
        )


class TheiaOutput:
    @staticmethod
    def all(*args: Input[T], **kwargs: Input[T]):
        if args and kwargs:
            raise ValueError(
                "TheiaOutput.all() was supplied a mix of named and unnamed inputs"
            )
        return Output.all(*args, **kwargs)

    def apply(
        self, func: Callable[[T_co], Input[U]], run_with_unknowns: Optional[bool] = None
    ) -> "Output[U]":
        return self.apply(func, run_with_unknowns)
